#ifndef __GRAPH_H__
#define __GRAPH_H__
#include <stddef.h>
#include "SList.h"

typedef struct _Graph Graph;

typedef int vertex;
typedef int weight;

typedef enum {
        GRAPH_TYPE_DIRECTED = 0x1,  /* if not directed, is assumed undirected */
        GRAPH_TYPE_DENSE    = 0x2,  /* if not dense, is assumed to be sparse */
        GRAPH_TYPE_WEIGHTED = 0x4,  /* edges have weights */
} GraphType;

/**
 * Creates a new graph.
 *
 * @param options specifies the type of graph to create.
 * @param number_of_vertices specifies the number of vertices the new graph will
 *       have. The vertices are numbered consecutively from 0 to
 *      (number_of_vertices - 1).
 */
Graph *graph_new                (GraphType options, size_t number_of_vertices);

/**
 * Destroys a graph and frees all of its allocated resources.
 */
void   graph_destroy            (Graph *graph);

/**
 * Adds a new edge to the graph.
 *
 * If the graph if weighted, the edge is added with the specified weight.
 * If the graph is un-weighted, the weight parameter is discarded, but its value
 * MUST be zero.
 *
 * @return whether the operation was successfull or not.
 */
int    graph_add_edge           (Graph *graph, vertex from_a, vertex to_b, weight weight);

/**
 * Tells whether the graph have and edge between two vertices.
 */
int    graph_exists_edge        (Graph *graph, vertex a, vertex b);

size_t graph_number_of_vertices (Graph *graph);
size_t graph_number_of_edges    (Graph *graph);

/**
 * Tells whether the graph is composed of one single strongly connected component.
 */
int    graph_is_connected       (Graph *graph);

/**
 * Returns an array with all the connected components of the graph.
 */
Graph** graph_get_components    (Graph *graph);

/**
 * Tells whether a vertex is on any path between two other vertices.
 */
int    graph_vertex_is_on_path  (Graph *graph, vertex v,
                                      vertex from_vertex, vertex to_vertex);

SList *graph_get_neighbours     (Graph *graph, vertex v);
void   graph_foreach_neighbour  (Graph *graph, vertex v, VisitorFunc visit, void *extra_data);

void   graph_foreach_vertex     (Graph *graph, VisitorFunc visit, void *extra_data);

/**
 * Visits each edge of the graph. If the graph is weighted, the 3rd argument to
 * TripletVisitor will be the weight, otherwise it will be NULL.
 */
void   graph_foreach_edge       (Graph *graph, TripletVisitor visit, void *extra_data);

/**
 * Traverses all the vertices of the connected component of the vertex 'initial'
 * using the DFS algorithm. It applies a visitor function to each visited vertex,
 * until the visitor function returns FALSE.
 *
 * When the visitor function returns FALSE, the traversal is aborted and control
 * flow returns to the caller.
 */
void   graph_dfs                (Graph *graph, vertex initial,
                                       CancellableVisitor visit, void *extra_data);

/**
 * Traverses all the vertices of the connected component of the vertex 'initial'
 * using the BFS algorithm. It applies a visitor function to each visited vertex,
 * until the visitor function returns FALSE.
 *
 * When the visitor function returns FALSE, the traversal is aborted and control
 * flow returns to the caller.
 */
void   graph_bfs                (Graph *graph, vertex initial,
                                       CancellableVisitor visit, void *extra_data);

/**
 * Returns a topological sort of the vertices of a directed graph, or NULL if
 * not such sorting exists.
 */
SList *graph_topological_sort   (Graph *graph);

#endif /* __GRAPH_H__ */
