#include <stddef.h>
#include "GraphAdjList.h"
#include "Common.h"
#include "Mem.h"
#include "Log.h"

#define GRAPH_ADJ_LIST_INITIAL_SIZE  128

typedef struct _GraphAdjListNode {
        vertex                   vertex;
        struct _GraphAdjListNode *next;
} GraphAdjListNode;

typedef struct _GraphAdjListWeightedNode {
        vertex                           vertex;
        weight                           weight;
        struct _GraphAdjListWeightedNode *next;
} GraphAdjListWeightedNode;

typedef struct {
        size_t       number_of_vertices;
        size_t       allocated_vertices;  /* number of entries in adjs array */
        union {
                GraphAdjListNode          **nodes;
                GraphAdjListWeightedNode  **weighted_nodes;
        } adjs;    /* array of adjacency lists */
        unsigned int is_directed : 1;
        unsigned int is_weighted : 1;
} GraphAdjList;

static GraphAdjListNode *graph_adj_list_node_new (vertex adjacent)
{
        GraphAdjListNode *adj_node = mem_new(GraphAdjListNode,1);
        adj_node->vertex = adjacent;
        adj_node->next = NULL;
        return adj_node;
}

/*
static
GraphAdjListWeightedNode *graph_adj_list_weighted_node_new (vertex adjacent, weight weight)
{
        GraphAdjListWeightedNode *adj_node = mem_new(GraphAdjListWeightedNode,1);
        adj_node->vertex = adjacent;
        adj_node->weight = weight;
        adj_node->next = NULL;
        return adj_node;
}
*/

static int
graph_adj_list_add_edge (void *_graph, vertex from_a, vertex to_b, weight weight)
{
        GraphAdjList *graph = (GraphAdjList*)_graph;
        GraphAdjListNode *node;

        log_return_val_if_fail(from_a < graph->number_of_vertices, FALSE);
        log_return_val_if_fail(to_b < graph->number_of_vertices, FALSE);

        if (graph->is_weighted) {
                /* TODO */                
        } else {
                if (graph->adjs.nodes[from_a] == NULL) {
                        graph->adjs.nodes[from_a] = graph_adj_list_node_new(to_b);
                } else {
                        for (node = graph->adjs.nodes[from_a]; node->next != NULL; node = node->next);
                        node->next = graph_adj_list_node_new(to_b);
                }
                if (!graph->is_directed) {
                        /* create the entry in the other direction */
                        if (graph->adjs.nodes[to_b] == NULL) {
                                graph->adjs.nodes[to_b] = graph_adj_list_node_new(from_a);
                        } else {
                                for (node = graph->adjs.nodes[to_b]; node->next != NULL; node = node->next);
                                node->next = graph_adj_list_node_new(from_a);
                        }
                }
        }

        return TRUE;
}

static int graph_adj_list_exists_edge (void *_graph, vertex a, vertex b)
{
        GraphAdjList *graph = (GraphAdjList*)_graph;
        GraphAdjListNode *node;
        GraphAdjListWeightedNode *weighted_node;

        if (graph->is_weighted) {
                for (weighted_node = graph->adjs.weighted_nodes[a];
                    weighted_node != NULL;
                    weighted_node = weighted_node->next) {
                        if (weighted_node->vertex == b) {
                                return TRUE;
                        }
                }
        } else {
                for (node = graph->adjs.nodes[a]; node != NULL; node = node->next) {
                        if (node->vertex == b) {
                                return TRUE;
                        }
                }
        }
        return FALSE;
}

static size_t graph_adj_list_number_of_vertices (void *_graph)
{
        GraphAdjList *graph = (GraphAdjList*)_graph;
        return graph->number_of_vertices;
}

static size_t graph_adj_list_number_of_edges (void *_graph)
{
        /* TODO */
        return 0;
}

static int graph_adj_list_is_connected (void *_graph)
{
        /* TODO */
        return FALSE;
}

static int
graph_adj_list_vertex_is_on_path (void *_graph, vertex v, vertex from, vertex to)
{
        /* TODO */
        return FALSE;
}

static GraphAdjList *graph_adj_list_new_internal (GraphType options, size_t number_of_vertices)
{
        size_t i;

        GraphAdjList *graph = mem_new(GraphAdjList, 1);
        graph->is_directed = options & GRAPH_TYPE_DIRECTED;
        graph->is_weighted = options & GRAPH_TYPE_WEIGHTED;
        if (graph->is_weighted) {
                fprintf(stderr, "DEBUG: creating weighted adjacency list\n");
                graph->adjs.weighted_nodes = mem_new(GraphAdjListWeightedNode*, number_of_vertices);
                for (i = 0; i < number_of_vertices; i++) {
                        graph->adjs.weighted_nodes[i] = NULL;
                }
        } else {
                fprintf(stderr, "DEBUG: creating un-weighted adjacency list\n");
                graph->adjs.nodes = mem_new(GraphAdjListNode*, number_of_vertices);
                for (i = 0; i < number_of_vertices; i++) {
                        graph->adjs.nodes[i] = NULL;
                }
        }
        graph->allocated_vertices = number_of_vertices;
        graph->number_of_vertices = number_of_vertices;
        return graph;
}

static void graph_adj_list_destroy (void *_graph)
{
        GraphAdjList *graph = (GraphAdjList*)_graph;
        GraphAdjListNode *node, *node_next;
        GraphAdjListWeightedNode *weighted_node, *weighted_node_next;
        int i;

        for (i = 0; i < graph->number_of_vertices; i++) {
                if (graph->is_weighted) {
                        weighted_node = graph->adjs.weighted_nodes[i];
                        while (weighted_node != NULL) {
                            weighted_node_next = weighted_node->next;
                            mem_free(weighted_node);
                            weighted_node = weighted_node_next;
                        }
                } else {
                        node = graph->adjs.nodes[i];
                        while (node != NULL) {
                            node_next = node->next;
                            mem_free(node);
                            node = node_next;
                        }
                }
        }
        if (graph->is_weighted) {
                mem_free(graph->adjs.weighted_nodes);
        } else {
                mem_free(graph->adjs.nodes);
        }
        mem_free(graph);
}

GraphImpl *graph_adj_list_new (GraphType options, size_t number_of_vertices)
{
        GraphImpl *graphImpl = mem_new(GraphImpl, 1);
        graphImpl->add_edge = graph_adj_list_add_edge;
        graphImpl->number_of_vertices = graph_adj_list_number_of_vertices;
        graphImpl->number_of_edges = graph_adj_list_number_of_edges;
        graphImpl->exists_edge = graph_adj_list_exists_edge;
        graphImpl->is_connected = graph_adj_list_is_connected;
        graphImpl->vertex_is_on_path = graph_adj_list_vertex_is_on_path;
        graphImpl->private_data = graph_adj_list_new_internal(options, number_of_vertices);
        graphImpl->destroy = graph_adj_list_destroy;
        return graphImpl;
}
